@@ -24,7 +24,7 @@ from calculate.core.server.gen_pid import search_worked_process
 from calculate.core.setup_cache import Cache as SetupCache
 from calculate.core.server.func import MethodsInterface
 from calculate.lib.cl_template import SystemIni, LayeredIni
-from calculate.lib.datavars import DataVarsError, VariableError
+from calculate.lib.datavars import DataVarsError, VariableError, Variable
 
 from calculate.lib.utils.tools import AddonError
 from calculate.lib.utils.colortext.palette import TextState
@@ -32,11 +32,16 @@ from calculate.lib.utils.colortext import get_color_print
 from calculate.update.emerge_parser import RevdepPercentBlock
 from calculate.update.datavars import DataVarsUpdate
 from calculate.update.update_info import UpdateInfo
-from calculate.lib.utils.binhosts import fetch_packages
+from calculate.lib.utils.binhosts import (Binhosts, BinhostSignError,
+    BinhostError, PackagesIndex, DAYS)
+from calculate.lib.utils.gpg import GPG, GPGError
 from calculate.lib.cl_log import log
+import hashlib
 import re
 import shutil
 from collections import MutableSet
+from contextlib import contextmanager
+import tempfile
 
 from calculate.lib.utils.git import Git, GitError, MTimeKeeper, NotGitError
 from calculate.lib.utils.portage import (Layman, EmergeLog,
@@ -51,7 +56,8 @@ Colors = TextState.Colors
 from calculate.lib.utils.files import (getProgPath, STDOUT, removeDir,
                                        PercentProgress, process, getRunCommands,
                                        readFile, listDirectory, pathJoin,
-                                       writeFile)
+                                       find, FindFileType,quite_unlink,
+                                       writeFile, makeDirectory)
 import emerge_parser
 import logging
 from emerge_parser import (EmergeParser, EmergeCommand, EmergeError,
@@ -149,6 +155,8 @@ class Update(MethodsInterface):
         self.refresh_binhost = False
         self.pkgnum = None
         self.pkgnummax = None
+        self.gpgdata_md5 = []
+        self.gpg_changed = False
 
     def get_prog_path(self, program_name):
         return getProgPath(program_name)
@@ -156,6 +164,30 @@ class Update(MethodsInterface):
     def getGit(self):
         return self.clVars.Get('cl_update_git')
 
+    @contextmanager
+    def private_repo(self, rpath, url):
+        if Git.is_private_url(url):
+            try:
+                if not path.exists(rpath):
+                    makeDirectory(rpath)
+                os.chmod(rpath, 0700)
+                yield
+            finally:
+                try:
+                    for dn in (Git._gitDir(rpath), path.join(rpath, "profiles/templates")):
+                        if path.exists(dn):
+                            os.chmod(dn, 0700)
+                    for fn in find(path.join(rpath, "profiles"), True, FindFileType.RegularFile,
+                                   True, None, downfilter=lambda x: not x.endswith("/templates")):
+                        if fn.endswith("calculate.env") or fn.endswith("ini.env"):
+                            os.chmod(fn, 0600)
+                    if path.exists(rpath):
+                        os.chmod(rpath, 0755)
+                except OSError:
+                    pass
+        else:
+            yield
+
     def _syncRepository(self, name, url, rpath, revision,
                         cb_progress=None, clean=False, notask=False):
         """
@@ -174,8 +206,9 @@ class Update(MethodsInterface):
                     self.startTask(_("Syncing the {rep} repository").format(
                         rep=name.capitalize()))
                 self.addProgress()
-                git.cloneTagRepository(url, rpath, revision,
-                                       cb_progress=cb_progress)
+                with self.private_repo(rpath, url):
+                    git.cloneTagRepository(url, rpath, revision,
+                                           cb_progress=cb_progress)
                 info_outdated = True
             else:
                 cr = ""
@@ -190,16 +223,17 @@ class Update(MethodsInterface):
                         status = git.getStatusInfo(rpath)
                         if not status or status['files']:
                             need_update = True
-                except GitError:
+                except GitError as e:
                     need_update = True
                 if need_update:
                     if not notask:
                         self.startTask(_("Syncing the {rep} repository").format(
                             rep=name.capitalize()))
                     self.addProgress()
-                    git.updateTagRepository(url, rpath, revision,
-                                            cb_progress=cb_progress,
-                                            clean=clean)
+                    with self.private_repo(rpath, url):
+                        git.updateTagRepository(url, rpath, revision,
+                                                cb_progress=cb_progress,
+                                                clean=clean)
                     new_cr = git.getCurrentCommit(rpath)
                     if new_cr != cr:
                         info_outdated = True
@@ -208,6 +242,8 @@ class Update(MethodsInterface):
                 dv.Set('cl_update_outdate_set', 'on', force=True)
         finally:
             self.unstash_cache(rpath, name)
+        # TODO: debug1
+        #dv.Set('cl_update_outdate_set', 'on', force=True)
         return True
 
     def raiseOutdate(self):
@@ -385,12 +421,6 @@ class Update(MethodsInterface):
             if repname not in ("portage", "gentoo"):
                 layman.add(repname, url, rpath_orig)
         finally:
-            try:
-                if (path.exists(rpath) and
-                            Git.parse_url(url)[0] == Git.GitProtocol.SSH):
-                    os.chmod(rpath, 0700)
-            except OSError:
-                pass
             mtime.restore()
         return True
 
@@ -1031,6 +1061,7 @@ class Update(MethodsInterface):
     def reconfigureProfileVars(self, profile_dv, chroot):
         """
         Синхронизировать репозитории
+        Настройка переменных для выполнения синхронизации репозиториев
         """
         dv = self.clVars
 
@@ -1052,8 +1083,9 @@ class Update(MethodsInterface):
                 dv.Set(var_name, profile_dv.Get(var_name), force=True)
             dv.Set('cl_chroot_path', chroot, force=True)
         except DataVarsError as e:
-            print str(e)
-            raise UpdateError(_("Wrong profile"))
+            error = UpdateError(_("Wrong profile"))
+            error.addon = e
+            raise error
         return True
 
     def setProfile(self, profile_shortname):
@@ -1191,28 +1223,23 @@ class Update(MethodsInterface):
         cache.update(force=True)
         return True
 
-    def check_binhost(self, write_binhost=True):
+    def get_bin_cache_filename(self):
+        return pathJoin(self.clVars.Get('cl_chroot_path'),
+                        LayeredIni.IniPath.Grp)
+
+    def update_local_info_binhost(self, write_binhost=True):
         """
         Проверить, что доступен хотя бы один из binhost'ов
         :return:
         """
         hosts = self.clVars.Get("update.cl_update_binhost_host")
         datas = self.clVars.Get("update.cl_update_binhost_revisions")
-        bin_cache_fn = pathJoin(self.clVars.Get('cl_chroot_path'),
-                                LayeredIni.IniPath.Grp)
         if not hosts:
-            self.clVars.Delete('cl_update_binhost', location="system")
-            try:
-                if path.exists(bin_cache_fn):
-                    os.unlink(bin_cache_fn)
-            except OSError:
-                raise UpdateError(
-                    _("Failed to remove cached ini.env of binary repository"))
-
+            self.delete_binhost()
             raise UpdateError(_("Failed to find the server with "
                                 "appropriate updates"))
         else:
-            with writeFile(bin_cache_fn) as f:
+            with writeFile(self.get_bin_cache_filename()) as f:
                 f.write(datas[0].strip()+"\n")
         if write_binhost:
             if hosts[0] != self.clVars.Get('update.cl_update_binhost'):
@@ -1225,9 +1252,9 @@ class Update(MethodsInterface):
                 if new_ts.isdigit():
                     ini = SystemIni(self.clVars)
                     ini.setVar('system', {'last_update': new_ts})
-        if self.clVars.Get('cl_action') == 'sync':
-            value = self.clVars.GetBool('cl_update_binhost_stable_set')
-            new_value = self.clVars.GetBool('cl_update_binhost_stable_opt_set')
+        if self.is_update_action(self.clVars.Get("cl_action")):
+            value = self.clVars.GetBool('update.cl_update_binhost_stable_set')
+            new_value = self.clVars.GetBool('update.cl_update_binhost_stable_opt_set')
             if value != new_value:
                 self.clVars.Write(
                     'cl_update_binhost_stable_set',
@@ -1235,6 +1262,9 @@ class Update(MethodsInterface):
                     location="system")
         return True
 
+    def is_update_action(self, action):
+        return action == 'sync'
+
     def save_with_bdeps(self):
         oldval = self.clVars.Get('cl_update_with_bdeps_set')
         newval = self.clVars.Get('cl_update_with_bdeps_opt_set')
@@ -1247,39 +1277,65 @@ class Update(MethodsInterface):
     def message_binhost_changed(self):
         if self.refresh_binhost:
             self.printWARNING(_("Update server was changed to %s") %
-                              self.clVars.Get('cl_update_binhost'))
+                              self.clVars.Get('update.cl_update_binhost'))
+            self.clVars.Set("update.cl_update_package_cache_set",
+                            Variable.On, force=True)
         else:
             self.printSUCCESS(_("Update server %s") %
-                              self.clVars.Get('cl_update_binhost'))
+                              self.clVars.Get('update.cl_update_binhost'))
         return True
 
-    def update_binhost_list(self):
+    def delete_binhost(self):
+        self.clVars.Delete('cl_update_binhost', location="system")
+        try:
+            bin_cache_fn = self.get_bin_cache_filename()
+            if path.exists(bin_cache_fn):
+                os.unlink(bin_cache_fn)
+        except OSError:
+            raise UpdateError(
+                _("Failed to remove cached ini.env of binary repository"))
+        try:
+            for varname in ('update.cl_update_package_cache',
+                            'update.cl_update_package_cache_sign'):
+                fn = self.clVars.Get(varname)
+                if path.exists(fn):
+                    os.unlink(fn)
+        except OSError:
+            raise UpdateError(
+                _("Failed to remove cached Package index"))
+        # удалить binhost
+        binhost_fn = self.inchroot(
+            self.clVars.Get("update.cl_update_portage_binhost_path"))
+        if path.exists(binhost_fn):
+            os.unlink(binhost_fn)
+        return False
+
+    def update_binhost_list(self, dv=None):
         """
         Обновить список binhost'ов после обновления до master веток
         :return:
         """
-        dv = DataVarsUpdate()
+        changes = False
         try:
-            dv.importUpdate()
-            dv.flIniFile()
-            changes = False
+            if dv is None:
+                dv = DataVarsUpdate()
+                dv.importUpdate()
+                dv.flIniFile()
             for varname in ('update.cl_update_binhost_list',
                             'update.cl_update_binhost_unstable_list',
                             'update.cl_update_binhost_timestamp_path',
-                            'cl_update_binhost_revision_path'):
+                            'update.cl_update_binhost_revision_path'):
                 new_value = dv.Get(varname)
                 old_value = self.clVars.Get(varname)
                 if new_value != old_value:
                     changes = True
-                self.clVars.Set(varname, new_value, force=True)
-            if not changes:
-                self.clVars.Delete('cl_update_binhost', location="system")
-                raise UpdateError(_("Failed to find the server with "
-                                    "appropriate updates"))
-        except DataVarsError:
-            self.clVars.Delete('cl_update_binhost', location="system")
-            raise UpdateError(_("Failed to find the server with "
-                                "appropriate updates"))
+                    self.clVars.Set(varname, new_value, force=True)
+        except DataVarsError as e:
+            raise UpdateError(_("Failed to get values for binhost search"))
+
+        if not changes:
+            return False
+        self.create_binhost_data()
         return True
 
     def drop_binhosts(self, dv):
@@ -1294,8 +1350,273 @@ class Update(MethodsInterface):
         dv.Invalidate('update.cl_update_rep_rev')
         return True
 
-    def download_packages(self, url_binhost, packages_fn):
-        fetch_packages(url_binhost, packages_fn)
+    def inchroot(self, fn):
+        return pathJoin(self.clVars.Get("cl_chroot_path"), fn)
+
+    def prepare_gpg(self):
+        """
+        Получить объект для проверки подписи, либо получить заглушку
+        """
+        gpg_force = self.clVars.Get('cl_update_gpg_force')
+        gpg_keys = [self.inchroot(x)
+                    for x in self.clVars.Get('cl_update_gpg_keys')]
+        if gpg_force == "skip":
+            return True
+        gpgtmpdir = "/var/calculate/tmp/update"
+        if not path.exists(gpgtmpdir):
+            makeDirectory(gpgtmpdir)
+        gpg = GPG(tempfile.mkdtemp(dir=gpgtmpdir,
+                                   prefix="gpg-"))
+        for keyfn in gpg_keys:
+            if path.exists(keyfn):
+                try:
+                    key = readFile(keyfn)
+                    gpg.import_key(key)
+                except GPGError as e:
+                    self.printWARNING(_("Failed to load public keys from '%s' "
+                        "for signature checking") % keyfn)
+        if not gpg.count_public():
+            if gpg_force == "force":
+                raise UpdateError(_("Public keys for Packages signature checking not found"))
+            else:
+                return True
+        self.clVars.Set('update.cl_update_gpg', gpg, force=True)
         return True
 
+    def download_packages(self, url_binhost, packages_fn, packages_sign_fn, gpg):
+        quite_unlink(packages_fn)
+        orig_packages = Binhosts.fetch_packages(url_binhost)
+        try:
+            with writeFile(packages_fn) as f:
+                pi = PackagesIndex(orig_packages)
+                pi["TTL"] = str(30 * DAYS)
+                pi["DOWNLOAD_TIMESTAMP"] = str(int(time.time()))
+                pi.write(f)
+        except (OSError, IOError):
+            raise UpdateError(_("Failed to save Packages"))
+        self.endTask(True)
+        self.startTask(_("Check packages index signature"))
+        if not gpg:
+            self.endTask("skip")
+            self.clVars.Set('cl_update_package_cache_set', Variable.Off, force=True)
+            return True
+        try:
+            Binhosts.check_packages_signature(
+                url_binhost, orig_packages, gpg)
+            with writeFile(packages_sign_fn) as f:
+                f.write(Binhosts.fetch_packages_sign(url_binhost))
+        except BinhostSignError:
+            for fn in (packages_fn, packages_sign_fn):
+                if path.exists(fn):
+                    try:
+                        os.unlink(fn)
+                    except OSError:
+                        pass
+            self.clVars.Set("update.cl_update_bad_sign_set", Variable.On)
+            self.clVars.Set('update.cl_update_binhost_recheck_set', Variable.On)
+            self.clVars.Set('cl_update_package_cache_set', Variable.Off, force=True)
+            raise
+        return True
 
+    class Reason(object):
+        Success = 0
+        BadSign = 1
+        Outdated = 2
+        Skip = 3
+        Updating = 4
+        WrongBinhost = 5
+        BadEnv = 6
+        EnvNotFound = 7
+        SkipSlower = 8
+        UnknownError = 9
+
+
+        @staticmethod
+        def humanReadable(reason):
+            return {
+                Update.Reason.WrongBinhost: "FAILED (Wrong binhost)",
+                Update.Reason.Outdated: "OUTDATED",
+                Update.Reason.Updating: "UPDATING",
+                Update.Reason.BadEnv: "FAILED (Bad env)",
+                Update.Reason.EnvNotFound: "FAILED (Env not found)",
+                Update.Reason.UnknownError: "FAILED (Unknown error)",
+                Update.Reason.BadSign: "FAILED (Bad sign)",
+                Update.Reason.Skip: "SKIP",
+                Update.Reason.SkipSlower: "",
+                Update.Reason.Success: ""
+            }.get(reason,reason)
+
+    def _get_binhost_logger(self):
+        return log("binhost-scan.log",
+                   filename=pathJoin(
+                       self.clVars.Get('cl_chroot_path'),
+                       "/var/log/calculate/binhost-scan.log"),
+                   formatter="%(message)s")
+
+    def get_arch_machine(self):
+        return self.clVars.Get('os_arch_machine')
+
+    @variable_module("update")
+    def create_binhost_data(self):
+        dv = self.clVars
+        last_ts = dv.Get('cl_update_last_timestamp')
+        if dv.GetBool('cl_update_binhost_stable_opt_set'):
+            binhost_list = dv.Get('cl_update_binhost_list')
+        else:
+            binhost_list = dv.Get('cl_update_binhost_unstable_list')
+        self.binhosts_data = Binhosts(
+            # значение малозначимо, поэтому берётся из собирающей системы
+            dv.GetInteger('cl_update_binhost_timeout'),
+            dv.Get('cl_update_binhost_revision_path'),
+            dv.Get('cl_update_binhost_timestamp_path'),
+            last_ts, binhost_list,
+            self.get_arch_machine(),
+            gpg=dv.Get('update.cl_update_gpg'))
+        return True
+
+    @variable_module("update")
+    def _search_best_binhost(self, binhosts_data, stabilization):
+        if not self.clVars.Get('cl_ebuild_phase'):
+            logger = self._get_binhost_logger()
+        if logger:
+            logger.info(
+                "Started scan on: {date}, current timestamp: {ts}".format(
+                date=time.ctime(), ts=binhosts_data.last_ts))
+        retval = []
+        skip_check_status = False
+        actual_reason = self.Reason.UnknownError
+        for binhost in sorted(binhosts_data.get_binhosts(), reverse=True):
+            host = binhost.host
+            if not binhost.valid:
+                reason = self.Reason.WrongBinhost 
+            elif binhost.outdated:
+                reason = self.Reason.Outdated 
+            elif not skip_check_status:
+                status = binhost.status
+                if status is not binhosts_data.BinhostStatus.Success:
+                    errors = {
+                        binhosts_data.BinhostStatus.Updating: self.Reason.Updating,
+                        binhosts_data.BinhostStatus.BadEnv: self.Reason.BadEnv,
+                        binhosts_data.BinhostStatus.EnvNotFound: self.Reason.EnvNotFound 
+                    }
+                    reason = errors.get(status, self.Reason.UnknownError)
+                elif binhost.bad_sign:
+                    reason = self.Reason.BadSign
+                else:
+                    # SUCCESS
+                    if not binhost.downgraded or stabilization:
+                        host = "-> %s" % host
+                        reason = self.Reason.Success
+                    else:
+                        reason = self.Reason.Skip
+            elif binhost.downgraded:
+                reason = self.Reason.Skip
+            else:
+                reason = self.Reason.SkipSlower
+
+            if reason == self.Reason.Success:
+                retval.append([binhost.host, binhost.data,
+                            str(binhost.timestamp),
+                            str(binhost.duration)])
+                skip_check_status = True
+            if reason < actual_reason:
+                actual_reason = reason
+            logger.info("{host:<60} {speed:<7} {timestamp:<10} {reason}".format(
+                host=host, speed=float(binhost.duration) / 1000.0,
+                timestamp=binhost.timestamp,
+                reason=Update.Reason.humanReadable(reason)))
+        if not retval:
+            if actual_reason is self.Reason.BadSign:
+                raise UpdateError(_("Failed to find the reliable server with appropriate updates"))
+            elif actual_reason in (self.Reason.Outdated,
+                                   self.Reason.Skip,
+                                   self.Reason.Updating):
+                raise UpdateError(_("Failed to find the server with appropriate updates"))
+            else:
+                raise UpdateError(_("Failed to find the working server with updates"))
+        return retval
+        
+    def check_current_binhost(self, binhost_url):
+        """
+        Проверка текущего сервера обновлений на валидность
+        """
+        if not binhost_url in self.binhosts_data.binhost_list:
+            raise UpdateError(_("Current binhost is absent in list of update servers"))
+        binhost = self.binhosts_data.get_binhost(binhost_url)
+
+        if binhost.valid and not binhost.outdated and not binhost.downgraded:
+            if binhost.status == self.binhosts_data.BinhostStatus.Success:
+                self.clVars.Set('update.cl_update_binhost_data',
+                        [[binhost.host, binhost.data,
+                         str(binhost.timestamp),
+                         str(binhost.duration)]],
+                         force=True)
+                self.endTask()
+        else:
+            if not binhost.valid:
+                raise UpdateError(
+                    _("Current binhost {} is not valid").format(binhost_url))
+            elif binhost.outdated:
+                raise UpdateError(
+                    _("Current binhost {} is outdated").format(binhost_url))
+            elif binhost.downgraded:
+                raise UpdateError(
+                    _("Binary packages on the current binhost {} "
+                      "are older than local").format(binhost_url))
+        if self.binhosts_data.gpg:
+            packages_fn = self.clVars.Get('update.cl_update_package_cache')
+            packages_sign_fn = self.clVars.Get('update.cl_update_package_cache_sign')
+            if path.exists(packages_fn) and path.exists(packages_sign_fn):
+                packages_sign = readFile(packages_sign_fn)
+                pi = PackagesIndex(readFile(packages_fn))
+                pi.clean()
+                try:
+                    Binhosts.check_packages_signature(
+                        None, pi.get_value(), self.binhosts_data.gpg,
+                        sign=packages_sign)
+                    return True
+                except BinhostSignError:
+                    for fn in (packages_fn, packages_sign_fn):
+                        if path.exists(fn):
+                            try:
+                                os.unlink(fn)
+                            except OSError:
+                                pass
+            if binhost.bad_sign:
+                raise UpdateError(
+                    _("Current binhost {} has wrong signature").format(
+                        binhost_url))
+        return True
+
+    @variable_module("update")
+    def detect_best_binhost(self):
+        # выполняется переход с серверов unstable обновлней на stable
+        # в этом случае не важно, что бинари могут старее текущих
+        if (self.clVars.GetBool('cl_update_binhost_stable_opt_set') and
+                not self.clVars.GetBool('cl_update_binhost_stable_set')):
+            stabilization = True
+        else:
+            stabilization = False
+
+        self.startTask(_("Searching new binhost"))
+        retval = self._search_best_binhost(self.binhosts_data, stabilization)
+
+        self.clVars.Set('update.cl_update_binhost_data',
+            retval or Variable.EmptyTable, force=True)
+
+        self.endTask()
+        return True
+
+    def update_layman(self):
+        """
+        Обновить базу layman
+        :param builder_path:
+        :return:
+        """
+        cmd = "/usr/bin/layman"
+        cmd_path = self.get_prog_path(cmd)
+        if not cmd_path:
+            raise UpdateError(_("Failed to find the %s command") % cmd)
+        layman = emerge_parser.CommandExecutor(cmd_path, ["-f"])
+        layman.execute()
+        return layman.success()
